// This file was generated by ecm_qt_declare_logging_category(): DO NOT EDIT!

#include "loggingcategory.h"


#if QT_VERSION >= QT_VERSION_CHECK(5, 4, 0)
Q_LOGGING_CATEGORY(KArchiveLog, "kf5.karchive", QtWarningMsg)
#else
Q_LOGGING_CATEGORY(KArchiveLog, "kf5.karchive")
#endif

